#include <iostream>
#include <sstream>
#include <limits>

// for std::pair
#include <utility>

// for std::sqrt
#include <cmath>

// for std::terminate
#include <exception>

#include <optional>

using namespace std;

class SqrtError {};

// Compute square root using Newton's method
double Sqrt(const double a) {
  if (a < 0) {
    throw SqrtError{};
  }
  
  if (a == 0) {
    return 0;
  }
  
  // calculate 2^(log₂ a) as the initial approximation
  double x = a;
  for (double t = a; t > 1; t /= 4, x /= 2) {}
  for (double t = a; t < 1; t *= 4, x *= 2) {}
  
  // 8 iterations are enough for doubles at this point
  for (int i = 0; i < 8; ++i) {
    // xᵢ₊₁ = xᵢ - f(xᵢ)/f'(xᵢ)
    //
    // where f(x) = x² - a, so the roots are +/- √a
    //       f'(x) = 2x
    x = (x + (a / x) ) / 2;
  }
  return x;
}

struct QuadraticError {
  virtual std::string message() const = 0;
};

struct NotQuadraticEqn : public QuadraticError {
  virtual std::string message() const override { return "the input is not a quadratic eqn"; }
};

struct NoRealRoots : public QuadraticError {
  double discriminant;

  NoRealRoots(double d) : discriminant(d) {}
  
  virtual std::string message() const override {
    return "the equation doesn't have real roots.";
  }
};

// Find the roots of given quadratic equation
pair<double, double> SolveQuadraticEqn(double a, double b, double c) {
  if (a == 0) {
    throw NotQuadraticEqn{};
  }
  // square root of discriminant

  // what if sqrt fails?
  double d = b * b - 4 * a * c;
  try {
    double r = Sqrt(d);
    return { (-b + r) / (2 * a), (-b - r) / (2 * a) };
  } catch (SqrtError) {
    throw NoRealRoots{d};
  }
}

void SolveAndPrint(double a, double b, double c) {
  try {
    auto [ root1, root2 ] = SolveQuadraticEqn(a, b, c);
    cout << "roots of " << a << "*x² + " << b << "*x + " << c << " = 0 are "
         << root1 << " and " << root2 << "\n";
  } catch (NotQuadraticEqn&) {
    if (b != 0) {
      auto root = - c / b;
      cout << "the root of " << a << "*x² + " << b << "*x + " << c << " = 0 is "
           << root << "\n";
    } else {
      cout << "eqn doesn't have roots\n";
    }
  } catch (NoRealRoots& err) {
    cout << "could not solve the equation "
         << a << "*x² + " << b << "*x + " << c << " = 0\n";
    cout << "the discriminant is " << err.discriminant << " < 0\n";
  }
  cout << "solver done\n";
}

string optToDouble(optional<double> o) {
  if (o) {
    stringstream s;
    s << *o;
    return s.str();
  }
  return "<no value>";
}

int main(int argc, char * argv[]) {
  // try {
  // 
  // cout << "sqrt 2 = " << Sqrt(2) << "\n";
  // cout << "sqrt 4 = " << Sqrt(4) << "\n";
  // cout << "sqrt 1/2 = " << Sqrt(0.5) << "\n";
  // cout << "sqrt 3/2 = " << Sqrt(1.5) << "\n";
  // cout << "sqrt 1/4 = " << Sqrt(0.5) << "\n";
  // cout << "sqrt 1024 = " << Sqrt(1024) << "\n";
  // cout << "sqrt 225e10 = " << Sqrt(225e10) << "\n";
  // cout << "sqrt 2 = " << sqrt(2) << "\n";
  // cout << "sqrt 4 = " << sqrt(4) << "\n";
  // cout << "sqrt 1/2 = " << sqrt(0.5) << "\n";
  // cout << "sqrt 3/2 = " << sqrt(1.5) << "\n";
  // cout << "sqrt 1/4 = " << sqrt(0.5) << "\n";
  // cout << "sqrt 1024 = " << sqrt(1024) << "\n";
  // cout << "sqrt 225e10 = " << sqrt(225e10) << "\n";
  // // TODO: error checking ↓
  // cout << "sqrt -3 = " << Sqrt(-3) << "\n";

  // } catch(SqrtError & err) { // const char * err) {
  //   cerr << "square root error\n";
  // }

  // cout << "DONE\n";

  if (argc == 4) {
    double a = stod(argv[1]);
    double b = stod(argv[2]);
    double c = stod(argv[3]);
    // TODO: error handling ↓
    auto [ root1, root2 ] = SolveQuadraticEqn(a, b, c);
    cout << "x1 = " << root1 << ", x2 = " << root2 << "\n";
    // cout << "could not solve the eqn: " << err.what() << "\n";
  }

  cout << "example equations:\n";
  SolveAndPrint(1, 0, -4);
  SolveAndPrint(1, 4, 4);
  SolveAndPrint(1, 5, 6);
  SolveAndPrint(1, 0, 3);
  SolveAndPrint(1, -5, 6);
  SolveAndPrint(1, -5, 4);
  SolveAndPrint(1, 0, 4);
  SolveAndPrint(0, 1, 4);
  SolveAndPrint(0, 0, 4);

  return 0;
}
